/*
** Copyright (c) 2002
**      Jeff Forys (jeff@forys.cranbury.nj.us).  All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that: (1) Redistributions of
** source code must retain the above copyright notice, this list of
** conditions and the following disclaimer, (2) Redistributions in
** binary form must reproduce the above copyright notice, this list
** of conditions and the following disclaimer in the documentation
** and/or other materials provided with the distribution, (3) All
** advertising materials mentioning features or use of this software
** must display the following acknowledgment: ``This product includes
** software developed by Jeff Forys (jeff@forys.cranbury.nj.us).'', (4)
** The name of the author may not be used to endorse or promote products
** derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

const char *Version = "1.2 20020214";
const char *CopyrightVersion = "camlgr %s\n\nCopyright (c) 2002 Jeff Forys.  All rights reserved.\n\nRedistribution and use in source and binary forms, with or without\nmodification, is permitted under the terms specified in the source\ncode (and manual pages).\n";

#ifndef lint
static char rcsid[] = "$Id: camlgr.c,v 1.8 2002/02/09 19:31:09 forys Exp $";
#endif

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <syslog.h>
#include <pwd.h>

#include "config.h"

int openfil(const char *);
int usage(const char *, const char *);
int version(void);

/*
 * camlgr - Apache httpd "piped logger".
 *
 * This program reads from STDIN and attempts to write to a file/fifo.
 * When a log message is read, if it turns out no one is reading from
 * the fifo (i.e., the open() or write() would block), the log message
 * is simply discarded.
 *
 * camlgr [-dv] [ <file> ]
 */
int
main(int argc, char *argv[])
{
	struct passwd *pw;
	struct sigaction act;
	const char *outfil;
	char *progname, buf[8192];
	int n, dbgflg = 0, fd = -1;

	/* if invoked as root, drop to CAMOWN */
	if ((pw = getpwnam(CAMOWN)) == NULL)		/* sigh, not fatal */
		syslog(LOG_WARNING, "getpwnam: no such user: %s", CAMOWN);
	else
		setuid(pw->pw_uid);

	SETPROGNAME(progname, argv);

	/* parse flags, return usage() error if not defined */
	while ((n = getopt(argc, argv, "dv")) != -1)
		switch (n) {
			case 'd':
				dbgflg = 1;
				break;
			case 'v':
				return version();
			case '?':
			default:
				return usage(progname, "unknown flag");
		}
	argc -= optind;
	argv += optind;

	/* check for file, return usage() error if anything is wrong */
	if (argc == 1)
		outfil = argv[0];
	else
		return usage(progname, "expected name of one log file");

	/* open log file */
	openlog(progname, LOG_NDELAY | LOG_PID, SYSLOGFAC);

	/* ignore the SIGPIPE (if fifo, and reader closes) */
	act.sa_flags = 0;
	act.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &act, (struct sigaction *)NULL);

	/* open output file (in case a reader is spinning on open) */
	fd = openfil(outfil);

	/*
	 * Loop continuously, always accepting input.  The assumption
	 * is that whoever started us will restart us if we go down.
	 * The most important task then, is to read from STDIN.
	 *
	 * We may be writing to a FIFO.  There may not be a reader.
	 * The open(2) on the FIFO may block; ensure that it doesnt.
	 * Set non-blocking on the descriptor so writes do not block.
	 * Be prepared for failures; drop data when there's no reader.
	 *
	 * If openfil() fails, 'fd' may be -2: never to write again!
	 */
	for(;;) {
		/* wait for data on STDIN */
		switch (n = read(STDIN_FILENO, buf, sizeof buf)) {
		    case -1:		/* error on read() */
			if (errno == EINTR)
				continue;
			syslog(LOG_ERR, "read(STDIN_FILENO): %m");
			return 2;

		    case 0:		/* EOF */
			if (dbgflg)
				syslog(LOG_DEBUG, "STDIN read() -> EOF");
			return 0;

		    default:		/* data read */
			/* if existing file not yet open, try (again) now */
			if (fd == -1)
				fd = openfil(outfil);
			/* if file is open, try to write (ignore errors) */
			if (fd >= 0) {
				int nwrite = write(fd, buf, n);
				if (dbgflg)
					syslog(LOG_DEBUG,
						"write(%d) -> %d (%d)",
						n, nwrite, errno);
			}
			break;
		}
	}
	/*NOTREACHED*/
}

/*
 * openfil(flnm) - try to open 'flnm' and make it non-blocking.
 *
 * Return file descriptor on success.  If file/device does not
 * exist, this may be transient so return -1 (leaving errno set).
 * If a permanent error occurs, syslog() an error and return -2.
 */
int
openfil(const char *flnm)
{
	int fl, fd = open(flnm, O_WRONLY|O_APPEND|O_NONBLOCK, 0600);

	if (fd < 0 && errno != ENXIO && errno != ENOENT) {
		syslog(LOG_ERR, "open(%s): %m", flnm);
		fd = -2;
	}

	/* must set non-blocking */
	if (fd >= 0 && ((fl = fcntl(fd, F_GETFL, 0)) < 0 ||
			fcntl(fd, F_SETFL, fl | O_NONBLOCK) < 0)) {
		syslog(LOG_ERR, "fcntl(O_NONBLOCK): %m");
		(void) close(fd);
		fd = -2;
	}

	return fd;
}

/*
 * usage(progname, errmsg) - print usage info to STDERR and return '1'.
 */
int
usage(const char *progname, const char *errmsg)
{
	fprintf(stderr, "%s: parse error: %s\n", progname, errmsg);
	fprintf(stderr, "%s [-dv] [<file>]\n", progname);
	return 1;
}

int
version(void)
{
	printf(CopyrightVersion, Version);
	return 0;
}
