/*
** Copyright (c) 2002
**      Jeff Forys (jeff@forys.cranbury.nj.us).  All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that: (1) Redistributions of
** source code must retain the above copyright notice, this list of
** conditions and the following disclaimer, (2) Redistributions in
** binary form must reproduce the above copyright notice, this list
** of conditions and the following disclaimer in the documentation
** and/or other materials provided with the distribution, (3) All
** advertising materials mentioning features or use of this software
** must display the following acknowledgment: ``This product includes
** software developed by Jeff Forys (jeff@forys.cranbury.nj.us).'', (4)
** The name of the author may not be used to endorse or promote products
** derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

/*
 * Compatibility routines.
 *
 * POSIX has left the building...
 */

#ifndef lint
static char rcsid[] = "$Id: compat.c,v 1.5 2002/01/12 20:18:27 forys Exp $";
#endif

#include <sys/param.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <limits.h>
#include <errno.h>

#include "compat.h"

#ifdef	HAS_BSDSTATFS

#include <sys/param.h>
#include <sys/mount.h>

/*
 * getmntpnt(dir)
 *
 * Return mount point for path specified in 'dir' or NULL on failure.
 *
 * This function depends on 4.4BSD statfs() and 'statfs.f_mntonname'.
 */
const char *
getmntpnt(const char *dir)
{
	static struct statfs sb;

	if (statfs(dir, &sb) < 0) {
		syslog(LOG_ERR, "statfs(%s): %m", dir);
		return NULL;
	}

	return sb.f_mntonname;
}

#else

#include <sys/vfs.h>
#ifdef sun
#include <sys/mntent.h>
#include <sys/mnttab.h>
#else
#include <mntent.h>
#endif

/*
 * getmntpnt(dir)
 *
 * Return mount point for path specified in 'dir' or NULL on failure.
 *
 * This function depends on one of the (many) varieties of getmntent().
 */
const char *
getmntpnt(const char *dir)
{
#if defined(sun)
	const char *mntfil = "/etc/mnttab";
	struct extmnttab mnt;
	int status;
#endif
#if defined(__hpux)
	const char *mntfil = "/etc/mnttab";
	struct mntent *mntp;
#endif
#if defined(__linux)
	const char *mntfil = "/etc/mtab";
	struct mntent *mntp;
#endif
	FILE *mtab;
	struct stat sb;
	dev_t dirdev;
	char *mntdir;

	if (stat(dir, &sb) < 0) {
		syslog(LOG_ERR, "stat(%s): %m", dir);
		return NULL;
	}
	dirdev = sb.st_dev;
	mntdir = NULL;

#ifdef sun
	if ((mtab = fopen(mntfil, "r")) == NULL) {
		syslog(LOG_ERR, "fopen(%s): failed", mntfil);
		return NULL;
	}
	resetmnttab(mtab);

	while ((status = getextmntent(mtab, &mnt, sizeof(mnt))) == 0) {
		if (stat(mnt.mnt_mountp, &sb) < 0)
			continue;
		if (sb.st_dev == dirdev) {
			mntdir = mnt.mnt_mountp;
			break;
		}
	}
	(void) fclose(mtab);
#else
	if ((mtab = setmntent(mntfil, "r")) == NULL) {
		syslog(LOG_ERR, "setmntent(%s): failed", mntfil);
		return NULL;
	}

	while ((mntp = getmntent(mtab)) != NULL) {
		if (stat(mntp->mnt_dir, &sb) < 0)
			continue;
		if (sb.st_dev == dirdev) {
			mntdir = mntp->mnt_dir;
			break;
		}
	}
	(void) endmntent(mtab);
#endif

	if (mntdir == NULL) {
		syslog(LOG_ERR, "getmntent(%s): cant find mount point", mntfil);
		return NULL;
	}

	return mntdir;
}
#endif

/* Linux declares flock() here */
#include <sys/file.h>

/*
 * exlock(fd, flnm)
 *
 * Acquire an exclusive lock on file descriptor 'fd' (file 'flnm')
 * using either flock() or fcntl().
 *
 * General note: If pid file is open across fork(), then it is locked.
 * Lock semantics may be flock() (lock held across fork) or fcntl()
 * (lock dropped across fork).  We *always* close the pid file after
 * fork() and never have multiple outstanding open descriptors on the
 * pid file; this makes the choice of lock semantics irrelevant.
 */
int
exlock(int fd, const char *flnm)
{
#ifdef	LOCK_EX
	if (flock(fd, LOCK_EX) < 0) {
		syslog(LOG_ERR, "flock(%s): %m", flnm);
		return 1;
	}
#else
	int ret;
	struct flock lockdes;
	int lock_cmd;
	int val;

	lockdes.l_whence = 0;
	lockdes.l_start = 0;
	lockdes.l_len = 0;
	lockdes.l_type = F_WRLCK;
	lock_cmd = F_SETLKW;

	if (fcntl(fd, F_SETLKW, &lockdes) < 0) {
		syslog(LOG_ERR, "fcntl(F_SETLKW,%s): %m", flnm);
		return 1;
	}
#endif
	return 0;
}

/*
 * removedir(dir)
 *
 * Remove directory 'dir' (and anything therein).
 *
 * Return 0 if 'dir' is gone, 1 if it seems to still be there.
 */
int
removedir(const char *dir)
{
	struct stat sb;
	char cmdbuf[PATH_MAX + 64];

	snprintf(cmdbuf, sizeof cmdbuf, "/bin/rm -rf %s", dir);
	system(cmdbuf);

	return (stat(dir, &sb) < 0 && errno == ENOENT)? 0: 1;
}
