/*
** Copyright (c) 2002
**      Jeff Forys (jeff@forys.cranbury.nj.us).  All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that: (1) Redistributions of
** source code must retain the above copyright notice, this list of
** conditions and the following disclaimer, (2) Redistributions in
** binary form must reproduce the above copyright notice, this list
** of conditions and the following disclaimer in the documentation
** and/or other materials provided with the distribution, (3) All
** advertising materials mentioning features or use of this software
** must display the following acknowledgment: ``This product includes
** software developed by Jeff Forys (jeff@forys.cranbury.nj.us).'', (4)
** The name of the author may not be used to endorse or promote products
** derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

#ifndef lint
static char rcsid[] = "$Id: parser.c,v 1.8 2002/02/09 19:12:21 forys Exp $";
#endif

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "parser.h"

static int			getstr(char **, char **, char **);
static int			getint(char **, int *, char **);
static int			getcam(char **, void *, char **);
static int			getcap(char **, void *, char **);
static int			addimg(char *, struct confcam_s *, char **);
static struct confcam_s *	camslot(char *, char **);

static struct confdata_s cf;

static struct parser_s {
	char *keyword;
	void *addr;
	int (*getarg)();
} parser[] = {
	{ "CAMDIR", &cf.camdir, getstr },
	{ "CURIMGDIR", &cf.curimgdir, getstr },
	{ "DEFIMGDIR", &cf.defimgdir, getstr },
	{ "CURCONDIR", &cf.curcondir, getstr },
	{ "CLIAPP", &cf.cliapp, getstr },
	{ "PIDFIL", &cf.pidfil, getstr },
	{ "LOGFIL", &cf.logfil, getstr },
	{ "CAMUSERS", &cf.camusers, getstr },
	{ "USERTOUT", &cf.usertout, getint },
	{ "CAPIVAL", &cf.capival, getint },
	{ "CAPDELAY", &cf.capdelay, getint },
	{ "SAMPLING", &cf.sampling, getint },
	{ "CAMERA", NULL, getcam },
	{ "CAPTURE", NULL, getcap },
	{ NULL }
};

static const int def_usertout = 30;	/* http client "gone" after 30 secs */
static const int def_capival = 1000;	/* 1 sec between each capture */
static const int def_capdelay = 0;	/* no delay to initial image capture */
static const int def_sampling = 0;	/* disable sampling */

/*
 * confinit()
 *
 * (Re)initialize confcam_s structure (cf), free any allocated memory.
 */
void
confinit(void)
{
	struct confcam_s *cfc;
	int i;

#define	_RESETSTR(s)	do { if ((s) != NULL) { free(s); s = NULL; } } while (0)
	_RESETSTR(cf.camdir);
        _RESETSTR(cf.curimgdir);
        _RESETSTR(cf.defimgdir);
        _RESETSTR(cf.curcondir);
        _RESETSTR(cf.cliapp);
        _RESETSTR(cf.pidfil);
        _RESETSTR(cf.logfil);
        _RESETSTR(cf.camusers);
        cf.usertout = cf.capival = cf.capdelay = cf.sampling = -1;

	for (cfc = cf.camera; cfc < &cf.camera[CONFCAMMAX]; cfc++) {
		_RESETSTR(cfc->name);
		_RESETSTR(cfc->capcmd);
		for (i = 0; i < CONFIMGMAX; i++)
			_RESETSTR(cfc->imgfil[i]);
		cfc->imgfil_e = cfc->imgfil;
	}
	cf.camera_e = cf.camera;
#undef	_RESETSTR
}

#define	ERROUT(m,str)	do { m = str; goto errout; } while (0)
#define	SKIPSPC(c)	while (isspace(*c)) c++

/*
 * confread(flnm, errmsg)
 *
 * Parse configuration file into confcam_s structure (cf).
 * Return pointer to structure on success, NULL on failure
 * (with '*errmsg' set to error message).
 */
struct confdata_s *
confread(const char *flnm, const char **errmsg)
{
	static char msgbuf[CONFMSGLEN];
	struct parser_s *parsp;
	char line[CONFLINELEN], *cp, *cp_e, *msg;
	FILE *fp;
	int len, lineno = 0;

	/* (re)initialize config data structure: free memory, reset vars */
	confinit();

	/* (re)open configuration file in preparation for parsing */
	if ((fp = fopen(flnm, "r")) == NULL)
		ERROUT(msg, "can not open file");

	/* parse each line in the configuration file */
	for (++lineno; fgets(line, sizeof line, fp) != NULL; lineno++) {
		if (*(cp = line) == '#')	/* comment line (ignored) */
			continue;
		SKIPSPC(cp);
		if (*cp == '\0')		/* blank line (ignored) */
			continue;
		for (cp_e = cp; *cp_e != '\0' && !isspace(*cp_e); cp_e++)
			;			/* advance past keyword */

		/* identify keyword and parse the first parameter */
		len = cp_e - cp;
		for (parsp = parser; parsp->keyword != NULL; parsp++) {
			if (strncasecmp(cp, parsp->keyword, len) == 0) {
				SKIPSPC(cp_e);
				if (*cp_e == '\0')
					ERROUT(msg, "missing parameter");
				if ((*parsp->getarg)(&cp_e, parsp->addr, &msg))
					goto errout;
				break;
			}
		}
		if (parsp->keyword == NULL)
			ERROUT(msg, "unknown keyword");

		SKIPSPC(cp_e);
		if (*cp_e != '\0' && *cp_e != '#')
			ERROUT(msg, "too many parameters");
	}

	/* initialize default values */
	if (cf.usertout == -1)
		cf.usertout = def_usertout;
	if (cf.capival == -1)
		cf.capival = def_capival;
	if (cf.capdelay == -1)
		cf.capdelay = def_capdelay;
	if (cf.sampling == -1)
		cf.sampling = def_sampling;
	if (cf.camusers == NULL) {
		if ((cf.camusers = malloc(3)) == NULL)
			ERROUT(msg, "out of memory");
		strcpy(cf.camusers, "::");
	}

	/* check required fields */
#define	_CHKSTR(str, errmsg)	if (str == NULL) ERROUT(msg, errmsg)
	_CHKSTR(cf.camdir, "CAMDIR not defined");
	_CHKSTR(cf.curimgdir, "CURIMGDIR not defined");
	_CHKSTR(cf.pidfil, "PIDFIL not defined");
#undef	_CHKSTR

	/* success: return new configuration data */
	fclose(fp);
	*errmsg = NULL;
	return &cf;

errout:
	/* failure: return NULL and explain the problem in 'errmsg' */
	if (fp != NULL)
		fclose(fp);
	confinit();			/* free any allocated memory on error */
	snprintf(msgbuf, sizeof msgbuf, "%s: line %d: %s", flnm, lineno, msg);
	*errmsg = msgbuf;
	return NULL;
}

/*
 * getstr(line, strp, errmsg)
 *
 * Read a string into newly allocated memory at '*strp'.
 * Handle escape sequences in string along the way.
 *
 * Return 0 on success, 1 on error ('*errmsg' set to error message).
 * On success, advance '*line' to next character in input line.
 */
static int
getstr(char **line, char **strp, char **errmsg)
{
	char str[CONFLINELEN];
	char *scp, *dcp, ch;

	if (*strp != NULL) {
		*errmsg = "keyword multiply defined";
		return 1;
	}

	/* string must start with a double quote */
	if (*(scp = *line) != '"') {
		*errmsg = "expected string parameter";
		return 1;
	}

	/* copy the string to 'str' handling any escape sequences */
	for (scp++, dcp = str; ; scp++, dcp++) {
		if (*scp == '\0') {
			*errmsg = "unterminated string parameter";
			return 1;
		}
		if ((*dcp = *scp) == '\\' && (ch = *(scp + 1)) != '\0') {
			switch(ch) {
			    case 'a': ch = '\007'; break;
			    case 'b': ch = '\010'; break;
			    case 'f': ch = '\014'; break;
			    case 'n': ch = '\012'; break;
			    case 'r': ch = '\015'; break;
			    case 't': ch = '\011'; break;
			    case 'v': ch = '\013'; break;
			    case '"': ch = '"'; break;
			    case '\'': ch = '\''; break;
			    case '\\': ch = '\\'; break;
			    default:
				*errmsg = "undefined character escape sequence";
				return 1;
			}
			*dcp = ch;
			scp++;
			continue;
		}
		if (*scp == '"') {
			*dcp = '\0';
			break;
		}
	}

	if (*str == '\0') {
		*errmsg = "empty strings are not permitted";
		return 1;
	}

	/* allocate memory for 'str' and copy the processed string */
	if ((*strp = malloc(strlen(str) + 1)) == NULL) {
		*errmsg = "out of memory allocating string";
		return 1;
	}
	strcpy(*strp, str);
	*line = scp + 1;
	return 0;
}

/*
 * getint(line, intp, errmsg)
 *
 * Read an integer into '*intp'.
 *
 * Return 0 on success, 1 on error ('*errmsg' set to error message).
 * On success, advance '*line' to next character in input line.
 */
static int
getint(char **line, int *intp, char **errmsg)
{
	char *int_e;

	if (!isdigit(**line)) {
		*errmsg = "expected numeric parameter";
		return 1;
	}
	*intp = strtol(*line, &int_e, 10);
	*line = int_e;
	return 0;
}

/*
 * getcam(line, unused, errmsg)
 *
 * Parse rest of line to read image files.
 *
 * Return 0 on success, 1 on error ('*errmsg' set to error message).
 * On success, advance '*line' to next character in input line.
 */
static int
getcam(char **line, void *unused, char **errmsg)
{
	struct confcam_s *cfc;
	char *cp = *line, *camnam = NULL, *vcp;

	/* read string containing camera name and find it's entry */
	if (getstr(&cp, &camnam, errmsg) == 1 ||
			(cfc = camslot(camnam, errmsg)) == NULL)
		return 1;

	/* ensure that camera names are sensible (e.g., underscores are bad) */
	for (vcp = camnam; *vcp != '\0'; vcp++)
		if (!isalnum(*vcp))
			break;
	if (*vcp != '\0') {
		*errmsg = "illegal camera name (letters and numbers only)";
		return 1;
	}

	/* read image files until out of space, comment, or end of line */
	for (;;) {
		SKIPSPC(cp);
		if (*cp == '\0' || *cp == '#')
			break;
		camnam = NULL;
		if (getstr(&cp, &camnam, errmsg) == 1 ||
				addimg(camnam, cfc, errmsg) == 1)
			return 1;
	}

	*line = cp;
	return 0;
}

/*
 * getcap(line, unused, errmsg)
 *
 * Parse rest of line to read capture command.
 *
 * Return 0 on success, 1 on error ('*errmsg' set to error message).
 * On success, advance '*line' to next character in input line.
 */
static int
getcap(char **line, void *unused, char **errmsg)
{
	struct confcam_s *cfc;
	char *cp = *line, *camnam = NULL;

	/* read string containing camera name and fine it's entry */
	if (getstr(&cp, &camnam, errmsg) == 1 ||
			(cfc = camslot(camnam, errmsg)) == NULL)
		return 1;

	if (cfc->capcmd != NULL) {
		*errmsg = "multiple capture commands";
		return 1;
	}

	/* skip any white space between camera name and capture command */
	SKIPSPC(cp);
	if (*cp == '\0') {
		*errmsg = "expected capture command";
		return 1;
	}

	/* get the capture command */
	if (getstr(&cp, &cfc->capcmd, errmsg) == 1)
		return 1;

	*line = cp;
	return 0;
}

/*
 * addimg(imgfil, cfc, errmsg)
 *
 * Add 'imgfil' to 'cfc->imgfil[].  Free 'imgfil' on error.
 *
 * Return 0 on success, 1 on error ('*errmsg' set to error message).
 * On success, advance '*line' to next character in input line.
 */
static int
addimg(char *imgfil, struct confcam_s *cfc, char **errmsg)
{
	int i;

	/* enforce CONFIMGLEN to prevent buffer overruns later on */
	if (strlen(imgfil) >= CONFIMGLEN)
		imgfil[CONFIMGLEN-1] = '\0';

	/* ensure 'imgfil' is not in cfc->imgfil[], then add it (if room) */
	for (i = 0; i < CONFIMGMAX; i++) {
		if (cfc->imgfil[i] == NULL) {
			cfc->imgfil[i] = imgfil;
			cfc->imgfil_e = &cfc->imgfil[i+1];
			break;
		}
		if (strcmp(imgfil, cfc->imgfil[i]) == 0) {
			free(imgfil);
			*errmsg = "multiply defined image file";
			return 1;
		}
	}

	if (i < CONFIMGMAX)			/* added */
		return 0;

	*errmsg = "image count exceeded (raise CONFIMGMAX)";
	return 1;
}

/*
 * camslot(camnam, errmsg)
 *
 * Locate 'camnam' in cf.camera[], create a new entry if necessary.
 * Free 'camnam' if duplicate or on error.
 *
 * Return pointer to cf.camera[] member on success, NULL if no space
 * left in array (with '*errmsg' set to error message).
 * On success, advance '*line' to next character in input line.
 */
static struct confcam_s *
camslot(char *camnam, char **errmsg)
{
	int i;

	/* find 'camnam' in cf.camera[], add it if not found (and if room) */
	for (i = 0; i < CONFCAMMAX; i++) {
		if (cf.camera[i].name == NULL) {
			cf.camera[i].name = camnam;
			cf.camera_e = &cf.camera[i+1];
			break;
		}
		if (strcmp(camnam, cf.camera[i].name) == 0) {
			free(camnam);
			break;
		}
	}

	if (i < CONFCAMMAX)			/* found or added */
		return &cf.camera[i];

	*errmsg = "camera count exceeded (raise CONFCAMMAX)";
	return NULL;
}

#ifdef	STANDALONE	/* enabled for stand-alone testing */
int
main()
{
	const char *errmsg;
	struct confdata_s *cfd;
	struct confcam_s *cfc;
	char **icp;

	if ((cfd = confread("cammgr.conf", &errmsg)) == NULL) {
		fprintf(stderr, "Error: %s\n", errmsg);
		return 1;
	}

	printf("CAMDIR <%s>\nCURIMGDIR <%s>\nDEFIMGDIR <%s>\nCLIAPP <%s>\n",
		cfd->camdir, cfd->curimgdir, cfd->defimgdir, cfd->cliapp);
	printf("CURCONDIR <%s>\nPIDFIL <%s>\nLOGFIL <%s>\nCAMUSERS <%s>\n",
		cfd->curcondir, cfd->pidfil, cfd->logfil, cfd->camusers);
	printf("USERTOUT <%d>\nCAPIVAL <%d>\nCAPDELAY <%d>\nSAMPLING <%d>\n",
		cfd->usertout, cfd->capival, cfd->capdelay, cfd->sampling);
	for (cfc = cfd->camera; cfc < cfd->camera_e; cfc++) {
		printf("CAMERA <%s> FILES:", cfc->name);
		for (icp = cfc->imgfil; icp < cfc->imgfil_e; icp++)
			printf(" <%s>", *icp);
		if (cfc->capcmd != NULL)
			printf("\nCAPTURE <%s>", cfc->capcmd);
		puts("");
	}

	return 0;
}
#endif
